;	dspfft1024.asm
;	
;				hacked together by Avery Wang -- April 10, 1992
;				Institut fuer Neuroinformatik
;				Ruhr-Uni-Bochum
;
;	This code was developed in a manner similar to the way wasp larvae
;	feed on their paralyzed prey--eating it from the inside out and
;	converting its biomass into its own--similar to what happens to
;	the astronauts on "Alien".  This used to be the dsp_dma_stream 
;	example given in /NextDeveloper/Examples/DSP
;
;----------------------------------------------------------------------
; DSP monitor illustrating DMA in and out of the DSP.
; J.O. Smith, 10/17/90
;
	page 255,255,0,1,1	   	; Width, height, topmar, botmar, lmar
	include 'ioequ.asm'

;--- Program configuration ---

BUG56_VERSION		set 0	; Set 0 for real thing, 1 for standalone Bug56
DEBUG	 		set 0	; Set to 0 after DSP code is debugged
DOUBLE_BUFFER		set 0	; 0 => simple&slow dma, 1 => double-buffered dma

	if BUG56_VERSION||DEBUG
START		equ 	$80	; Leaves room for Bug56's degmon monitor
	else
START		equ 	$40	; Leaves room for interrupt vectors
	endif

;--- DMA configuration ---

		define	R_DMA_IN 'R6'	; Dedicated to DMA in
		define	N_DMA_IN 'N6'	; Unused
		define	M_DMA_IN 'M6'	; -1

		define	R_DMA_OUT 'R7'	; Dedicated to DMA out
		define	N_DMA_OUT 'N7'	; Unused
		define	M_DMA_OUT 'M7'	; -1

Y_DMASTAT	equ	$AC05	; Off-chip Y memory word used for DMA status
Y_SYSCALL	equ	$AC06	; Off-chip Y memory word used for syscall arg
	if DOUBLE_BUFFER
DMA_READ_SIZE	equ	$0400	; Size of each DMA transfer in 
DMA_WRITE_SIZE	equ	$0400	; Size of each DMA transfer out
READ_BUF1	equ	$3000	; First input buffer
READ_BUF2	equ	$3400	; Second input buffer
WRITE_BUF1	equ	$3800	; First output buffer
WRITE_BUF2	equ	$3C00	; Second output buffer
	else
LENGTH			equ $0400
DMA_READ_SIZE	equ	LENGTH	; Size of each half-DMA transfer in (Real & Imag)
DMA_WRITE_SIZE	equ	LENGTH	; Size of each half-DMA transfer out (Real & Imag)
XFRM_SIZE		equ LENGTH	; Size of FFT
READ_BUF1		equ	$A000	; Input buffer
WRITE_BUF1		equ	$A400	; Output buffer
SINCOS_TAB		equ $A800	; sine and cosine table
FFT_DATA_IN		equ $A000	; obvious what this is.
FFT_DATA_OUT	equ $A400	; obvious what this is.
	endif

	include 'sincos.asm'
	sincos 	XFRM_SIZE,SINCOS_TAB	; set up sine, cosine table

;-------------------------

	if BUG56_VERSION

; Make handy symbol names for the Y control variables
; and do NOT assemble the reset vector for Bug56 version.

def_y_sym macro y_sym
	org y:y_sym
l_\y_sym dc 0
	endm
	
	def_y_sym Y_DMASTAT
	def_y_sym Y_SYSCALL

	else 

	org p:0
	jmp reset

	endif
	
	org p:START
;
; reset - first thing executed when DSP boots up - must be on-chip
;
reset	movec   #6,omr			; Data rom enabled, mode 2 = "normal"
	movec   #0,sp			; Clear stack
	bset    #0,x:m_pbc		; Enable host port
	bset	#3,x:m_pcddr		;    pc3 asserts 0 to enable external
	bclr	#3,x:m_pcd		;    DSP ram on very early machines
	movep   #>$000000,x:m_bcr	; No wait states for the external sram
    movep   #>$00B400,x:m_ipr  	; Intr levels: SSI=2, SCI=1, HOST=0
	jsr 	dma_reset		; Reset DMA state machine
	bset    #m_hcie,x:m_hcr		; enable host command interrupts
	move	#0,sr			; go to lowest int. priority level
	jmp 	main
;
; input/output routines, DMA and non-DMA i/o
;
	if DOUBLE_BUFFER
	  include 'fast_dma_support.asm'
	else
	  include 'slow_dma_support.asm'
	endif
	
;
; macro files: fftr2en.asm,sincos.asm
;
	include 'fftr2en.asm'
	include 'sign_extend.asm'
	include 'normalize.asm'
;
; main -- read scale factor, then loop forever scaling input signal
;
main

loop
	jsr	dma_buf_in
	jcs dma_stop			; flush and halt on abort signal
	sign_extend	FFT_DATA_IN,LENGTH
	fftr2en	FFT_DATA_IN,SINCOS_TAB,FFT_DATA_OUT
	normalize FFT_DATA_OUT,LENGTH,@log(XFRM_SIZE)/@log(2.)

	jsr dma_buf_out
	jmp loop

	end reset

	if *>$200
		warn 'Internal P memory overflow!' ; fatal for 8K SRAM
	endif
