;
;	DMA streaming support.  Cannibalized from simple_dma_support.asm by
;	J O Smith.
;------------------------- DSP <--> host communication ----------------------------
;
SC_W_REQ	equ	$020002	  ;"Sys call" requesting DMA write on chan 2
DM_R_REQ	equ	$050001	  ;"DSP message" requesting DMA read on chan 1
DM_W_REQ	equ	$040002	  ;message requesting DMA write on channel 2 
VEC_R_DONE	equ	$0024	  ;host command indicating dma read complete
VEC_W_DONE	equ	$0028	  ;host command indicating dma write complete
VEC_SYS_CALL	equ	$002C	  ;host command indicating sys-call int coming
VEC_ABORT 	equ 	$002E	  ;host command indicating end of input data
;
;------------------------- Bit fields in the status word ----------------------
;
B_DMA_DONE	equ	0	  ;indicates dma complete
B_SYS_CALL	equ	1	  ;indicates sys call has been received
B_ABORTING	equ	2	  ;indicates abort host command received
B_LAST_W_BUF	equ	3	  ;set after abort during last input buf
B_IDLE   	equ	4	  ;abort complete (looked at using Bug56)
;
;------------------------------- Interrupt vectors ----------------------------
;
dma_s_saved_lc	set *

	org	p:VEC_R_DONE
	bset	#B_DMA_DONE,y:Y_DMASTAT
	nop

	org	p:VEC_W_DONE
	bset	#B_DMA_DONE,y:Y_DMASTAT
	nop

	org	p:VEC_SYS_CALL
	jsr	>sys_call

	org	p:VEC_ABORT
 	bset 	#B_ABORTING,y:Y_DMASTAT
	nop

	org	p:dma_s_saved_lc

; sys_call - field a request from the kernel
; A "system call" is a host command followed by one int written to the DSP.
; In the future, the int may specify more ints to follow.
; arg = 24bits = (8,,16) = (op,datum)
; 	where op = 1 for read and 2 for write
;	and datum = 1 for "software fix" needed for DMA chips less than rev. 313
;		The fix tosses the first 16 bytes of the DMA write (to the DSP)
;		into the bit bucket.
;
sys_call
	jclr	#m_hrdf,x:m_hsr,sys_call	;buzz until int received
	movep	x:m_hrx,y:Y_SYSCALL		;int specifying operation
	bset	#B_SYS_CALL,y:Y_DMASTAT		;set flag to say we got this
	rti
	

	  ;; 
	  ;; Send a buffer to the host via "DSP-initiated DMA"
	  ;; Was putWordDMA--modified by Avery Wang July 21,1991
	  
dma_buf_out  move #WRITE_BUF1,X0			;reset to buffer start

	  bclr #B_DMA_DONE,y:Y_DMASTAT 	;reset end-of-dma indicator bit
	  move X0,R_DMA_OUT
	  move #-1,M_DMA_OUT
_htdeBuzz jclr	#m_htde,x:m_hsr,_htdeBuzz 	;wait until we can write host
	  movep	#DM_R_REQ,x:m_htx		;send "read request" DSP msg
_hf1Buzz  jclr	#m_hf1,x:m_hsr,_hf1Buzz 	;HF1 means DMA is set up to go

	; send Real values -- x memory
	  do #DMA_WRITE_SIZE,_dmaLoop_Re		;service the DMA transfer
_dmaBuzz_Re    jclr  #m_htde,x:m_hsr,_dmaBuzz_Re	;DMA hardware handshake
	    if DEBUG
	      jsclr #m_dma,x:m_hsr,dma_error	;reality check
	    endif
	    movep  x:(R_DMA_OUT)+,x:m_htx	;send values
_dmaLoop_Re

	; send Real values -- y memory
	  move	X0,R_DMA_OUT			;reset to start for refill
	  do #DMA_WRITE_SIZE,_dmaLoop_Im		;service the DMA transfer
_dmaBuzz_Im    jclr  #m_htde,x:m_hsr,_dmaBuzz_Im	;DMA hardware handshake
	    if DEBUG
	      jsclr #m_dma,x:m_hsr,dma_error	;reality check
	    endif
	    movep  y:(R_DMA_OUT)+,x:m_htx	;send values
_dmaLoop_Im

_doneBuzz  
		move y:Y_DMASTAT,y0
		jset	#B_DMA_DONE,y0,_endDMA ;terminating host command
	    jclr  #m_htde,x:m_hsr,_doneBuzz	;wait until we can write host
	    movep #0,x:m_htx			;send 0s until DMA completion 
	    jmp	  _doneBuzz
_endDMA   jset	#m_hf1,x:m_hsr,_endDMA		;make sure HF1 is low
	  move	X0,R_DMA_OUT			;reset to start for refill
	rts

;
;----end of routine ---
;

;check_abort - set carry bit if aborting and input is finished
;
check_abort
	jclr	#m_hf0,x:m_hsr,no_hf0		;old abort signal
 	bset 	#B_ABORTING,y:Y_DMASTAT
no_hf0 	btst 	#B_LAST_W_BUF,y:Y_DMASTAT	;aborted AND last buf read
	rts					;result in carry bit

;
;dma_stop - flush last buffer, signal to host and terminate execution.
;	    It is important that there be no pending W_REQ.
; 	    Due to bug #10138, sending an R_REQ while the driver is in
;	    penddmaout0 mode will cause the R_REQ to be ignored.  This
;	    of course is going to make the double-buffered, interrupt-driven
;	    I/O case a lot harder.
dma_stop
	bclr	#3,x:m_hcr		;clear HF2 to indicate idle state
	bset	#4,x:m_hcr		;set HF3 to indicate idle state
 	bset 	#B_IDLE,y:Y_DMASTAT	;for Bug56
	move	#>$FF,A	
	move	#WRITE_BUF1,x0			;reset to buffer start
	move	#-1,M0
	move	#-1,M4
	move	x0,R4
	move	x0,R0
	rep 	#DMA_WRITE_SIZE			;service the DMA transfer
	move	a,x:(R0)+	a,y:(R4)+
idle	jclr	#m_htde,x:m_hsr,idle	;buzz forever here until we can write
	
	jsr dma_buf_out		; until host has all it wants
	jmp	idle			;buzz



;; 
;; Get a buffer from the host via "DSP-initiated DMA"
;; warped by Avery Wang 7/21/91
dma_buf_in
	  move y:Y_DMASTAT,y0		; Kludged over by Avery Wang-6/23/91
	  jset  #B_LAST_W_BUF,y0,dma_in_done ;did last buf already
	  bclr	#B_DMA_DONE,y0		;reset end-of-dma indicator bit
	  bclr	#B_SYS_CALL,y0		;reset "sys-call pending" bit
	  move y0,y:Y_DMASTAT
_htdeBuzz jclr	#m_htde,x:m_hsr,_htdeBuzz 	;wait until we can talk to host
	  movep	#DM_W_REQ,x:m_htx		;send "write request" DSP msg
;;
;; Note: There is no way to abort the write.  The host MUST send the DMA
;; data after a W_REQ.  This makes the end-game harder.  When data for the DSP
;; runs out on the host side, the user is sent a "region completed" message.
;; However, by this time the DSP has normally requested the next input buffer,
;; and the driver is sitting in state penddmaout0, waiting for a region.
;; There is currently no API in the driver for aborting this state.  Therefore,
;; the user must hold the last DMA buffer's worth in a separate last region.
;; The DSP is sent the abort signal BEFORE the final region
;; is sent, and the DSP must not request any more input buffers.
;; Note that it is easier when the DSP is given an exact sample count at the
;; beginning, or if the DSP never terminates and the host keeps sending
;; buffers of zeros until it gets what it wants from the DSP.
;;
_hcBuzz	  jsr	check_abort			;check HF0
	  move y:Y_DMASTAT,y0
 	  jclr 	#B_ABORTING,y0,_no_abort
	  bset  #B_LAST_W_BUF,y0	;this is last input buffer
	  move y0,y:Y_DMASTAT
_no_abort move #SC_W_REQ,x0			;Expected sys_call
	  jclr #B_SYS_CALL,y0,_hcBuzz	;wait for "sys_call" host cmd
	  move y:Y_SYSCALL,b			; single int arg to host cmd
	  cmp x0,b				; see if it's what we expected
	  jne _hcBuzz				; if not, ignore it
	  bclr #B_SYS_CALL,y0			;got it
	  move y0,y:Y_DMASTAT

	  ; HF1 need not be checked since syscall implies DMA data follows
	  
	; load dma buffer into Real Space--x memory
	  move		#READ_BUF1,R_DMA_IN
	  move		#-1,M_DMA_IN
	  do #DMA_READ_SIZE,_dmaLoop_Re		;service the DMA transfer
_dmaBuzz_Re    jclr  #m_hrdf,x:m_hsr,_dmaBuzz_Re	;DMA hardware handshake
	    movep  x:m_hrx,x:(R_DMA_IN)+	;receive values
	    if DEBUG
	      jsclr #m_dma,x:m_hsr,dma_error	;reality check
	    endif		
_dmaLoop_Re

	; load dma buffer into Real Space--y memory
	  move		#READ_BUF1,R_DMA_IN
	  do #DMA_READ_SIZE,_dmaLoop_Im		;service the DMA transfer
_dmaBuzz_Im    jclr  #m_hrdf,x:m_hsr,_dmaBuzz_Im	;DMA hardware handshake
	    movep  x:m_hrx,y:(R_DMA_IN)+	;receive values
	    if DEBUG
	      jsclr #m_dma,x:m_hsr,dma_error	;reality check
	    endif
_dmaLoop_Im
	  move   #>READ_BUF1,x0			;read-buffer start address
	  move	x0,R_DMA_IN			;reset to start for consumption

_doneBuzz move y:Y_DMASTAT,y0
		jset	#B_DMA_DONE,y0,dma_in_done ;wait for terminating hc
	    jclr  #m_hrdf,x:m_hsr,_doneBuzz	;if DMA is still coming at us
	    movep x:m_htx,x0			;  receive it to drain its pipe
	    jmp	  _doneBuzz
dma_in_done
	btst  #B_LAST_W_BUF,y:Y_DMASTAT		;set carry bit appropriately
	rts
;
;----end of routine ---
;

;
; dma_reset - executed when DSP boots up
;
dma_reset
	move #READ_BUF1,R_DMA_IN	; Initialize input pointer
	move #WRITE_BUF1,R_DMA_OUT	; Initialize output pointer
	move #-1,M_DMA_IN		; Linear addressing
	move M_DMA_IN,M_DMA_OUT		;   for input and output
	move #0,X0
	move X0,y:Y_DMASTAT		; Clear DMA state
	rts
;
; Simple host-interface i/o
;
; "writeWordHost source" writes word in source to the host interface.
; "readWordHost dest" reads word in host interface to dest.
;

writeWordHost macro source
	if !BUG56_VERSION
_whbuzz	jclr	#m_htde,x:m_hsr,_whbuzz
	endif
        movep  	source,x:m_hrx
	endm	

readWordHost macro dest
	if !BUG56_VERSION
_rhbuzz	jclr	#m_hrdf,x:m_hsr,_rhbuzz
	endif
        movep  	x:m_hrx,dest
	endm	

	    if DEBUG
dma_error	ori #1,mr ; raise level to 1 (lock out host interrupts)
		bset	#m_hf3,x:m_hcr	; abort code = HF2 and HF3
		bset	#m_hf2,x:m_hcr
dma_abort	jmp 	dma_abort
	    endif
