; dspmemsize.asm [V0.90, 92-09-10]
;
; Written by Izumi Ohzawa.
; This program is in the public domain.
;
;; Program to find if the external DSP RAM is 8kw or 32kw.
;; It does not do a real test, but relies on memory map of 8k/32kword configurations.
;; If larger DSP RAM expansion becomes available for NeXT, this program will break and
;; must be modified accordingly.
;; Currently, it simply test if there are actually 32k unique memory cells that can
;; retain values.  This is done by writing the address value into each memory location,
;; and reading them back to see if the address == content.
;; If any mismatch is found, then we just assume that the external RAM size is 8kw.
;; Mismatches will be found for 8kw configuration because 8kw of RAM is aliased into
;; 32kw address space multiple times, i.e., same physical memory cell appears at
;; 4 different addresses.
;; If all 32kwords can retain unique values, we assume the RAM size is 32kw (it may be
;; actually larger if larger SIMM is installed.
;; Since all addresses are presented with memory cells, we don't have to worry about
;; floating data lines.
;;
	include 'ioequ.asm'

YC_Base		equ	$4000		;32kword contiguous Y memory starts here


	org	p:$0
	jmp	>reset
	
	org	p:$40

getSamp					;get from the host into a1
	jclr	#m_hrdf,x:m_hsr,getSamp
	movep	x:m_hrx,a
	rts

putSamp					;put a1 up to the host
	jclr	#m_htde,x:m_hsr,putSamp
	movep	a,x:m_htx
	rts

reset
	movec   #6,omr			;data rom enabled, mode 2
	bset    #0,x:m_pbc		;host port
	bset	#3,x:m_pcddr		;pc3 is an output with value
	bclr	#3,x:m_pcd		;zero to enable the external ram
	movep   #>$000000,x:m_bcr	;no wait states on the external sram


main
	jsr	getSamp			; wait for sample from host to start

; write the address of the location into memory cells

	move	#>YC_Base,R0
	move	#>$8000,X0		;32kwords to write
	do	X0,_storeAddress
	move	R0,Y1
	move	Y1,y:(R0)+		;write address via Y memory space
	nop
_storeAddress

; read it back and see if the address actually are stored in all 32kword locations

	clr	a
	move	#>YC_Base,R0
	move	#>$8000,X0		;32kwords to write
	do	X0,_readAddress
	move	R0,Y1			;current address in Y1
	move	y:(R0)+,a		;read the content of that address via Y memory space
	cmp	Y1,a			;do they match ?
	jne	_failed_match		;mismatch
_readAddress

; If you are here, all address locations were OK -- must be 32kwords
	move	#>$8000,a1
	jsr	putSamp
	jmp	stop

_failed_match				; must be 8kword
	move	#>$2000,a1
	jsr	putSamp
stop
	jmp	stop


