;;    GCC56K -- GCC 1.40 machine description for DSP56000/1 processor
;;    Copyright (C) 1991 Andrew Sterian and Bell-Northern Research
;;
;;    This program is free software; you can redistribute it and/or modify
;;    it under the terms of the GNU General Public License as published by
;;    the Free Software Foundation; either version 1, or (at your option)
;;    any later version.
;;
;;    This program is distributed in the hope that it will be useful,
;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;    GNU General Public License for more details.
;;
;;    You should have received a copy of the GNU General Public License
;;    along with the GCC distribution; if not, write to the Free Software
;;    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;;	  Contact the author at 'asterian@bnr.ca' or post to the 'comp.dsp'
;;	  newsgroup.

;;
;; Address register updates
;;
(define_insn ""
 [(set (match_operand:SI 0 "register_operand" "=C")
   (mem:SI (plus:HI (match_operand:HI 1 "register_operand" "a")
            (match_operand:HI 2 "general_operand" "g"))))]
 ""
 "*
{
	rtx xoperands[2];
	static char p[20];
	int reg1=REGNO(operands[1]);
	
	xoperands[0] = gen_rtx(REG, HImode, reg1+8);
	xoperands[1] = operands[2];
	output_asm_insn(\"move %s1,%0\", xoperands);
	sprintf(p, \"move %c:(%%1+N),%%0\", ((reg1==10) || (reg1==16)) ?
										 StackSpace : DataSpace);
    regValid[REGNO(operands[0])]=1;
	return p;
}")

;;
;; Special cases to handle address operations
;;
(define_insn ""
 [(set (match_operand:HI 0 "general_operand" "=m,r")
    (match_operand:SI 1 "general_operand" "A,mi"))]
 ""
 "*
{
	if (which_alternative == 1) {
		int regnum = REGNO(operands[0]);

        regValid[regnum] = 1;
		output_preloadN(operands[1]);
		if ((regnum==5) || (regnum==8)) {
			if (GET_CODE(operands[1]) == CONST_INT) {
				if (INTVAL(operands[1]) == 0) {
					return \"clr  %0\";
				}
			}
		}
	}
	output_preloadN(operands[0]);
	return \"move %w1,%0\";
}")

(define_insn ""
 [(set (match_operand 0 "general_operand" "=m,r,r")
    (subreg (match_operand:DI 1 "general_operand" "A,A,m") 1))]
 ""
 "*
{
	int reg0, reg1;

	switch(which_alternative) {
     case 0:	/* Register-to-memory */
		output_preloadN(operands[0]);
		return \"move %x1,%0\";

	 case 1:	/* Register-to-register */
		reg0 = REGNO(operands[0]); reg1 = REGNO(operands[1]);
		if ((reg0 < 4) && (reg0 == (reg1+1))) {
			return \";DI->HI*\";
		} else {
            regValid[reg0] = 1;
			return \"move %x1,%0\";
		}

	 case 2:	/* Memory-to-register */
		/* This regValid assignment must come before the call */
		/* to double_reg_from_memory since this routine may */
		/* invalidate a register. */
		regValid[REGNO(operands[0])]=1;
		if (TARGET_NOLSPACE) {
			double_reg_from_memory(operands);
			return \"\";
		} else {
			output_preloadN(operands[1]);
			return \"move Y:%Z1,%0\";
		}
	}
}")

;;
;; Instructions not needed by RTL
;;
;; MAC
(define_insn ""
 [(set (match_operand:SF 0 "register_operand" "=l,l")
   (plus:SF (match_operand:SF 1 "register_operand" "0,0")
    (mult:SF (match_operand:SF 2 "register_operand" "%d,e")
             (match_operand:SF 3 "register_operand" "e,d"))))]
 ""
 "*{ extendReg(operands[0]);
     return \"macr %2,%3,%0\";
}")

(define_insn ""
 [(set (match_operand:SI 0 "register_operand" "=l,l")
   (plus:SI (match_operand:SI 1 "register_operand" "0,0")
    (mult:SI (match_operand:SI 2 "register_operand" "%d,e")
             (match_operand:SI 3 "register_operand" "e,d"))))]
 ""
 "*{ extendReg(operands[0]);
     return \"asl  %0\;move %0,%x0\;mac  %2,%3,%0\;move %x0,%0\;asr  %0\";
}")

;; ADDL/ADDR
;; Don't know what to do if source/dest are the same

;;
;; Move instructions
;;
(define_insn "movqi"
 [(set (match_operand:QI 0 "general_operand" "=rm,r")
       (match_operand:QI 1 "general_operand" "r,mi"))]
 ""
 "*
{
	if (which_alternative == 1) {
		int regnum = REGNO(operands[0]);

        regValid[regnum]=1;
		output_preloadN(operands[1]);
		if ((regnum==5) || (regnum==8)) {
			if (GET_CODE(operands[1]) == CONST_INT) {
				if (INTVAL(operands[1]) == 0) {
					return \"clr  %0\";
				}
			}
		}
	} else {
		if (GET_CODE(operands[0]) == REG) {
			int reg0=REGNO(operands[0]);
			int reg1=REGNO(operands[1]);

			if ((reg0==5) || (reg0==8)) {
                if ((reg1==5) || (reg1==8))
                    regValid[reg0]=regValid[reg1];
                else
                    regValid[reg0]=1;
				if (reg1 < 9) {
					return \"tfr  %1,%0\";
				}
			}
		}
	}
	output_preloadN(operands[0]);
	return \"move %w1,%0\";
}")

(define_insn "movhi"
 [(set (match_operand:HI 0 "general_operand" "=rm,r")
       (match_operand:HI 1 "general_operand" "r,mi"))]
 ""
 "*
{
	if (which_alternative == 1) {
		int regnum = REGNO(operands[0]);

        regValid[regnum]=1;
		output_preloadN(operands[1]);
		if ((regnum==5) || (regnum==8)) {
			if (GET_CODE(operands[1]) == CONST_INT) {
				if (INTVAL(operands[1]) == 0) {
					return \"clr  %0\";
				}
			}
		}
	} else {
		if (GET_CODE(operands[0]) == REG) {
			int reg0=REGNO(operands[0]);
			int reg1=REGNO(operands[1]);

			if ((reg0==5) || (reg0==8)) {
                if ((reg1==5) || (reg1==8))
                    regValid[reg0]=regValid[reg1];
                else
                    regValid[reg0]=1;
				if ((reg1 < 4) || (reg1==5) || (reg1==8)) {
					return \"tfr  %1,%0\";
				}
			}
		}
	}
	output_preloadN(operands[0]);
	return \"move %w1,%0\";
}")

(define_insn "movsi"
 [(set (match_operand:SI 0 "general_operand" "=mA,A")
        (match_operand:SI 1 "general_operand" "A,mi"))]
 ""
 "*
{
	if (which_alternative == 1) {
		int regnum = REGNO(operands[0]);

        regValid[regnum]=1;
		output_preloadN(operands[1]);
		if ((regnum==5) || (regnum==8)) {
			if (GET_CODE(operands[1]) == CONST_INT) {
				if (INTVAL(operands[1]) == 0) {
					return \"clr  %0\";
				}
			}
		}
	} else {
		if (GET_CODE(operands[0]) == REG) {
			int reg0=REGNO(operands[0]);
			int reg1=REGNO(operands[1]);

			if ((reg0==5) || (reg0==8)) {
                if ((reg1==5) || (reg1==8))
                    regValid[reg0]=regValid[reg1];
				if ((reg1 < 4) || (reg1==5) || (reg1==8)) {
					return \"tfr  %1,%0\";
				}
			}
		}
	}
	output_preloadN(operands[0]);
	return \"move %w1,%0\";
}")

(define_insn "movdi"
 [(set (match_operand:DI 0 "general_operand" "=m,A,A,A,A")
       (match_operand:DI 1 "general_operand" "A,A,m,F,n"))]
 ""
 "*
{
	int intLow, intHigh;
	char *p;
	rtx xoperands[2];
	int reg0, reg1;

	switch(which_alternative) {
     case 0:	/* Register-to-memory */
		double_reg_to_memory(operands);
		return \"\";

	 case 1:	/* Register-to-register */
		reg0 = REGNO(operands[0]); reg1 = REGNO(operands[1]);
		if (((reg0==5) && (reg1==8)) || ((reg0==8) && (reg1==5))) {
            regValid[reg0]=regValid[reg1];
			return \"tfr  %1,%0\";
		}
		if (reg0 < 4) {	/* Destination is X or Y */
			return \"move %x1,%x0\;move %y1,%y0\";
		}
		if (reg1 < 4) { /* Destination is A or B, source is X or Y */
            regValid[reg0]=1;
			return \"move %y1,%0\;move %x1,%x0\";
		} /* else Destination is A or B, source is A or B */
        regValid[reg0]=regValid[reg1];
        return \"tfr  %1,%0\";

	 case 2:	/* Memory-to-register */
		double_reg_from_memory(operands);
        regValid[REGNO(operands[0])]=1;
		return \"\";

	 case 3:	/* ConstantDouble-to-register */
		intLow = CONST_DOUBLE_LOW(operands[1]);
		intHigh = CONST_DOUBLE_HIGH(operands[1]) << 8;
		intHigh |= (intLow >> 24);
		intLow &= 0xFFFFFF;
		
        regValid[REGNO(operands[0])]=1;
		xoperands[0] = operands[0];
		xoperands[1] = gen_rtx(CONST_INT, VOIDmode, intHigh);
		if (REGNO(operands[0]) < 4)
			p = \"move %1,%y0\";	/* Moves into X1, Y1 */
		else						/* Moves into A, B */
			if (intHigh == 0) 
				p = \"clr  %0\";
			else
				p = \"move %1,%0\";
		output_asm_insn(p, xoperands);

		xoperands[1] = gen_rtx(CONST_INT, VOIDmode, intLow);
		if (REGNO(operands[0]) > 3)
			if (intLow == 0)
				return \"\";

		output_asm_insn(\"move %1,%x0\", xoperands);
		return \"\";

	 case 4:	/* ConstantInt-to-register */
        regValid[REGNO(operands[0])]=1;
		if (REGNO(operands[0]) < 4) {
			output_asm_insn(\"move %1,%x0\", operands);
			return \"move #0,%y0\";
		} else {
			if (INTVAL(operands[1]) == 0)
				return \"clr  %0\";
			else
				return \"clr  %0\;move %1,%x0\";
		}
	}
}")

(define_insn "movsf"
 [(set (match_operand:SF 0 "general_operand" "=m,C,C,C")
       (match_operand:SF 1 "general_operand" "C,C,m,F"))]
 ""
 "*
{
	int intVal;
	rtx xoperands[2];

	switch(which_alternative) {
     case 0:	/* Register-to-memory */
		output_preloadN(operands[0]);
		return \"move %1,%0\";

	 case 1:	/* Register-to-register */
		if (REGNO(operands[0]) > 4) {
            if (REGNO(operands[1]) > 4)
                regValid[REGNO(operands[0])]=regValid[REGNO(operands[1])];
            else
                regValid[REGNO(operands[0])]=1;
			return \"tfr  %1,%0\";
		}
		return \"move %1,%0\";

	 case 2:	/* Memory-to-register */
		output_preloadN(operands[1]);
        regValid[REGNO(operands[0])]=1;
        return \"move %1,%0\";

	 case 3:	/* Constant-to-register */
		if (!const_double_to_SF(operands[1], &intVal))
			error(\"Illegal floating point value\");
		
        regValid[REGNO(operands[0])]=1;
		if ((intVal == 0) && (REGNO(operands[0]) > 3)) {
			return \"clr  %0\";
		}
		xoperands[1] = gen_rtx(CONST_INT, VOIDmode, intVal);
		xoperands[0] = operands[0];
		output_asm_insn(\"move %1,%0\", xoperands);
		return \"\";
	}
}")

(define_insn "movdf"
 [(set (match_operand:DF 0 "general_operand" "=m,A,A,A")
       (match_operand:DF 1 "general_operand" "A,A,m,F"))]
 ""
 "*
{
	int intLow, intHigh;
	rtx xoperands[2];
	char *p;
	int reg0, reg1;

	switch(which_alternative) {
     case 0:	/* Register-to-memory */
		double_reg_to_memory(operands);
		return \"\";

	 case 1:	/* Register-to-register */
		reg0 = REGNO(operands[0]); reg1 = REGNO(operands[1]);
		if (reg0 < 4) {	/* Destination is X or Y */
			return \"move %x1,%x0\;move %y1,%y0\";
		}
		if (reg1 < 4) { /* Destination is A or B, source is X or Y */
            regValid[REGNO(operands[0])]=1;
			return \"move %y1,%0\;move %x1,%x0\";
		} /* else Destination is A or B, source is A or B */
        regValid[REGNO(operands[0])]=regValid[REGNO(operands[1])];
		return \"tfr  %1,%0\";

	 case 2:	/* Memory-to-register */
		double_reg_from_memory(operands);
        regValid[REGNO(operands[0])]=1;
		return \"\";

	 case 3:	/* Constant-to-register */
		if (!const_double_to_DF(operands[1], &intHigh, &intLow)) 
			error(\"Illegal floating point value\");

        regValid[REGNO(operands[0])]=1;
		xoperands[0] = operands[0];
		xoperands[1] = gen_rtx(CONST_INT, VOIDmode, intHigh);
		if (REGNO(operands[0]) < 4)
			p = \"move %1,%y0\";	/* Moves into X1 or Y1 */
		else
			if (intHigh == 0)
				p = \"clr  %0\";
			else
				p = \"move %1,%0\";		/* Moves into A, B */
		output_asm_insn(p, xoperands);

		xoperands[1] = gen_rtx(CONST_INT, VOIDmode, intLow);
		if (REGNO(operands[0]) > 3)
			if (intLow == 0)
				return \"\";

		output_asm_insn(\"move %1,%x0\", xoperands);
		return \"\";
	}
}")

(define_insn "addqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
    (plus:QI (match_operand:QI 1 "register_operand" "%0")
             (match_operand:QI 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, ADD_56K); }")

(define_insn "addhi3"
 [(set (match_operand:HI 0 "register_operand" "=l,?a")
   (plus:HI (match_operand:HI 1 "register_operand" "0,a")
            (match_operand:HI 2 "general_operand" "C,g")))]
 ""
 "*{ if (which_alternative==0)
		return mathOpInsn(operands, ADD_56K);
	else {
	int same = (REGNO(operands[0]) == REGNO(operands[1]));
	int plus=0,minus=0;
	rtx xoperands[2];

	if (GET_CODE(operands[2]) == CONST_INT) {
		plus = (INTVAL(operands[2]) == 1);
		minus = (INTVAL(operands[2]) == -1);
	}
	if (plus)
		return (same ? \"move (%0)+\" : \"lua  (%1)+,%0\");
	if (minus)
		return (same ? \"move (%0)-\" : \"lua  (%1)-,%0\");
	xoperands[0] = gen_rtx(REG, HImode, REGNO(operands[1])+8);
	xoperands[1] = operands[2];
	output_asm_insn(\"move %s1,%0\", xoperands);
	return (same ? \"move (%0)+N\" : \"lua  (%1)+N,%0\");
	}
}")

;
; For address operations
;
(define_insn ""
 [(set (match_operand:HI 0 "register_operand" "=l")
   (plus:HI (match_operand:SI 1 "register_operand" "%0")
            (match_operand:SI 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, ADD_56K); }")

(define_insn "addsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (plus:SI (match_operand:SI 1 "register_operand" "%0")
            (match_operand:SI 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, ADD_56K); }")

(define_insn "adddi3"
 [(set (match_operand:DI 0 "register_operand" "=u")
   (plus:DI (match_operand:DI 1 "register_operand" "%0")
            (match_operand:DI 2 "register_operand" "A")))]
 ""
 "* { return mathOpInsn(operands, ADD_56K); }")

(define_insn "addsf3"
 [(set (match_operand:SF 0 "register_operand" "=l")
   (plus:SF (match_operand:SF 1 "register_operand" "%0")
            (match_operand:SF 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, ADD_56K); }")

(define_insn "adddf3"
 [(set (match_operand:DF 0 "register_operand" "=u")
   (plus:DF (match_operand:DF 1 "register_operand" "%0")
            (match_operand:DF 2 "register_operand" "A")))]
 ""
 "* { return mathOpInsn(operands, ADD_56K); }")

(define_insn "subqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
    (minus:QI  (match_operand:QI 1 "register_operand" "0")
            (match_operand:QI 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, SUB_56K); }")

(define_insn "subhi3"
 [(set (match_operand:HI 0 "register_operand" "=l,?a")
   (minus:HI   (match_operand:HI 1 "register_operand" "0,a")
            (match_operand:HI 2 "general_operand" "C,g")))]
 ""
 "*{ if (which_alternative==0)
		return mathOpInsn(operands, SUB_56K);
	 else {
	int same = (REGNO(operands[0]) == REGNO(operands[1]));
	int plus=0,minus=0;
	rtx xoperands[2];

	if (GET_CODE(operands[2]) == CONST_INT) {
		plus = (INTVAL(operands[2]) == 1);
		minus = (INTVAL(operands[2]) == -1);
	}
	if (plus)
		return (same ? \"move (%0)-\" : \"lua  (%1)-,%0\");
	if (minus)
		return (same ? \"move (%0)+\" : \"lua  (%1)+,%0\");
	xoperands[0] = gen_rtx(REG, HImode, REGNO(operands[1])+8);
	xoperands[1] = operands[2];
	output_asm_insn(\"move %s1,%0\", xoperands);
	return (same ? \"move (%0)-N\" : \"lua  (%1)-N,%0\");
	}
}")

;
; For address operations
;
(define_insn ""
 [(set (match_operand:HI 0 "register_operand" "=l")
   (minus:HI   (match_operand:SI 1 "register_operand" "0")
            (match_operand:SI 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, SUB_56K); }")

(define_insn "subsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (minus:SI   (match_operand:SI 1 "register_operand" "0")
            (match_operand:SI 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, SUB_56K); }")

(define_insn "subdi3"
 [(set (match_operand:DI 0 "register_operand" "=u")
   (minus:DI   (match_operand:DI 1 "register_operand" "0")
            (match_operand:DI 2 "register_operand" "A")))]
 ""
 "* { return mathOpInsn(operands, SUB_56K); }")

(define_insn "subsf3"
 [(set (match_operand:SF 0 "register_operand" "=l")
   (minus:SF   (match_operand:SF 1 "register_operand" "0")
            (match_operand:SF 2 "register_operand" "C")))]
 ""
 "* { return mathOpInsn(operands, SUB_56K); }")

(define_insn "subdf3"
 [(set (match_operand:DF 0 "register_operand" "=u")
   (minus:DF   (match_operand:DF 1 "register_operand" "0")
            (match_operand:DF 2 "register_operand" "A")))]
 ""
 "* { return mathOpInsn(operands, SUB_56K); }")

(define_insn "mulqi3"
 [(set (match_operand:QI 0 "register_operand" "=l,l")
   (mult:QI (match_operand:QI 1 "register_operand" "%d,e")
            (match_operand:QI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

(define_insn "mulhi3"
 [(set (match_operand:HI 0 "register_operand" "=l,l")
   (mult:HI (match_operand:HI 1 "register_operand" "%d,e")
            (match_operand:HI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

;
; For address operations
;
(define_insn ""
 [(set (match_operand:HI 0 "register_operand" "=l,l")
   (mult:HI (match_operand:SI 1 "register_operand" "%d,e")
            (match_operand:SI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

(define_insn "mulsi3"
 [(set (match_operand:SI 0 "register_operand" "=l,l")
   (mult:SI (match_operand:SI 1 "register_operand" "%d,e")
            (match_operand:SI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

;
; This exists explicitly for address operations
;
;(define_insn "muldi3"
; [(set (match_operand:DI 0 "register_operand" "=u")
;   (mult:DI (match_operand:DI 1 "register_operand" "%d")
;            (match_operand:DI 2 "register_operand" "d")))]
; "" Don't forget to set regValid
; "mpy  %x1,%x2,%0\;asr  %0")

(define_insn "mulsf3"
 [(set (match_operand:SF 0 "register_operand" "=l,l")
   (mult:SF (match_operand:SF 1 "register_operand" "%d,e")
            (match_operand:SF 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpyr %2,%1,%0\";
}")

(define_insn "umulqi3"
 [(set (match_operand:QI 0 "register_operand" "=l,l")
   (umult:QI (match_operand:QI 1 "register_operand" "%d,e")
            (match_operand:QI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

(define_insn "umulhi3"
 [(set (match_operand:HI 0 "register_operand" "=l,l")
   (umult:HI (match_operand:HI 1 "register_operand" "%d,e")
            (match_operand:HI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

;
; For address operations
;
(define_insn ""
 [(set (match_operand:HI 0 "register_operand" "=l,l")
   (umult:HI (match_operand:SI 1 "register_operand" "%d,e")
            (match_operand:SI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

(define_insn "umulsi3"
 [(set (match_operand:SI 0 "register_operand" "=l,l")
   (umult:SI (match_operand:SI 1 "register_operand" "%d,e")
            (match_operand:SI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

;
; This exists explicitly for address operations
;
;(define_insn "umuldi3"
; [(set (match_operand:DI 0 "register_operand" "=u")
;   (umult:DI (match_operand:DI 1 "register_operand" "%d")
;            (match_operand:DI 2 "register_operand" "d")))]
; "" Don't forget to set regValid
; "mpy  %x1,%x2,%0\;asr  %0")

(define_insn "andqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (and:QI  (match_operand:QI 1 "register_operand" "%0")
            (match_operand:QI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"and  %2,%0\";
}")

(define_insn "andhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (and:HI  (match_operand:HI 1 "register_operand" "%0")
            (match_operand:HI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"and  %2,%0\";
}")

(define_insn "andsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (and:SI  (match_operand:SI 1 "register_operand" "%0")
            (match_operand:SI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"and  %2,%0\";
}")

(define_insn "iorqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (ior:QI  (match_operand:QI 1 "register_operand" "%0")
            (match_operand:QI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"or   %2,%0\";
}")

(define_insn "iorhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (ior:HI  (match_operand:HI 1 "register_operand" "%0")
            (match_operand:HI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"or   %2,%0\";
}")

(define_insn "iorsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (ior:SI  (match_operand:SI 1 "register_operand" "%0")
            (match_operand:SI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"or   %2,%0\";
}")

(define_insn "xorqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (xor:QI  (match_operand:QI 1 "register_operand" "%0")
            (match_operand:QI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"eor  %2,%0\";
}")

(define_insn "xorhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (xor:HI  (match_operand:HI 1 "register_operand" "%0")
            (match_operand:HI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"eor  %2,%0\";
}")

(define_insn "xorsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (xor:SI  (match_operand:SI 1 "register_operand" "%0")
            (match_operand:SI 2 "register_operand" "d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"eor  %2,%0\";
}")

(define_insn "mulhisi3"
 [(set (match_operand:SI 0 "register_operand" "=l,l")
   (mult:HI (match_operand:HI 1 "register_operand" "%d,e")
            (match_operand:HI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

;
; For address operations
;
;(define_insn "mulhidi3"
; [(set (match_operand:DI 0 "register_operand" "=l,l")
;   (mult:HI (match_operand:HI 1 "register_operand" "%d,e")
;            (match_operand:HI 2 "register_operand" "e,d")))]
; "" Don't forget to set regValid
; "mpy  %2,%1,%0\;asr  %0")

(define_insn "mulqihi3"
 [(set (match_operand:HI 0 "register_operand" "=l,l")
   (mult:QI (match_operand:QI 1 "register_operand" "%d,e")
            (match_operand:QI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

(define_insn "mulsidi3"
 [(set (match_operand:DI 0 "register_operand" "=u,u")
   (mult:SI (match_operand:SI 1 "register_operand" "%d,e")
            (match_operand:SI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;asr  %0\";
}")

(define_insn "umulqihi3"
 [(set (match_operand:HI 0 "register_operand" "=l,l")
   (umult:QI (match_operand:QI 1 "register_operand" "%d,e")
            (match_operand:QI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

(define_insn "umulhisi3"
 [(set (match_operand:SI 0 "register_operand" "=l,l")
   (umult:HI (match_operand:HI 1 "register_operand" "%d,e")
            (match_operand:HI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;move %x0,%0\;asr  %0\";
}")

;
; For address operations
;
;(define_insn "umulhidi3"
; [(set (match_operand:DI 0 "register_operand" "=l,l")
;   (umult:HI (match_operand:HI 1 "register_operand" "%d,e")
;            (match_operand:HI 2 "register_operand" "e,d")))]
; "" Don't forget to set regValid
; "mpy  %2,%1,%0\;asr  %0")

(define_insn "umulsidi3"
 [(set (match_operand:DI 0 "register_operand" "=u,u")
   (umult:SI (match_operand:SI 1 "register_operand" "%d,e")
            (match_operand:SI 2 "register_operand" "e,d")))]
 ""
 "*{ regValid[REGNO(operands[0])]=1;
     return \"mpy  %2,%1,%0\;asr  %0\";
}")

(define_insn "ashlqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (ashift:QI (match_operand:QI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "*{ return shiftOpInsn(operands, ASHL_56K); }")

(define_insn "ashlhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (ashift:HI (match_operand:HI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "*{ return shiftOpInsn(operands, ASHL_56K); }")

;
; For address operations
;
(define_insn ""
 [(set (match_operand:HI 0 "register_operand" "=l")
   (ashift:HI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHL_56K); }")

(define_insn "ashlsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (ashift:SI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHL_56K); }")

(define_insn "ashldi3"
 [(set (match_operand:DI 0 "register_operand" "=u")
   (ashift:DI (match_operand:DI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHL_56K); }")

(define_insn "ashlsf3"
 [(set (match_operand:SF 0 "register_operand" "=l")
   (ashift:SF (match_operand:SF 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHL_56K); }")

(define_insn "ashldf3"
 [(set (match_operand:DF 0 "register_operand" "=u")
   (ashift:DF (match_operand:DF 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHL_56K); }")

(define_insn "ashrqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (ashiftrt:QI (match_operand:QI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHR_56K); }")

(define_insn "ashrhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (ashiftrt:HI (match_operand:HI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHR_56K); }")

;
; For address operations
;
(define_insn ""
 [(set (match_operand:HI 0 "register_operand" "=l")
   (ashiftrt:HI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHR_56K); }")

(define_insn "ashrsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (ashiftrt:SI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHR_56K); }")

(define_insn "ashrdi3"
 [(set (match_operand:DI 0 "register_operand" "=u")
   (ashiftrt:DI (match_operand:DI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHR_56K); }")

(define_insn "ashrsf3"
 [(set (match_operand:SF 0 "register_operand" "=l")
   (ashiftrt:SF (match_operand:SF 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHR_56K); }")

(define_insn "ashrdf3"
 [(set (match_operand:DF 0 "register_operand" "=u")
   (ashiftrt:DF (match_operand:DF 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ASHR_56K); }")

(define_insn "lshlqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (lshift:QI (match_operand:QI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, LSHL_56K); }")

(define_insn "lshlhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (lshift:HI (match_operand:HI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, LSHL_56K); }")

(define_insn "lshlsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (lshift:SI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, LSHL_56K); }")

(define_insn "lshrqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (lshiftrt:QI (match_operand:QI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, LSHR_56K); }")

(define_insn "lshrhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (lshiftrt:HI (match_operand:HI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, LSHR_56K); }")

(define_insn "lshrsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (lshiftrt:SI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, LSHR_56K); }")

(define_insn "rotlqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (rotate:QI (match_operand:QI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ROTL_56K); }")

(define_insn "rotlhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (rotate:HI (match_operand:HI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ROTL_56K); }")

(define_insn "rotlsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (rotate:SI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ROTL_56K); }")

(define_insn "rotrqi3"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (rotatert:QI (match_operand:QI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ROTR_56K); }")

(define_insn "rotrhi3"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (rotatert:HI (match_operand:HI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ROTR_56K); }")

(define_insn "rotrsi3"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (rotatert:SI (match_operand:SI 1 "register_operand" "0")
              (match_operand:SI 2 "general_operand" "rmJ")))]
 ""
 "* { return shiftOpInsn(operands, ROTR_56K); }")

(define_insn "negqi2"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (neg:QI (match_operand:QI 1 "register_operand" "0")))]
 ""
 "*{ extendReg(operands[0]);
     return \"neg  %0\";
}")

(define_insn "neghi2"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (neg:HI (match_operand:HI 1 "register_operand" "0")))]
 ""
 "*{ extendReg(operands[0]);
     return \"neg  %0\";
}")

(define_insn "negsi2"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (neg:SI (match_operand:SI 1 "register_operand" "0")))]
 ""
 "*{ extendReg(operands[0]);
     return \"neg  %0\";
}")

(define_insn "negsf2"
 [(set (match_operand:SF 0 "register_operand" "=l")
   (neg:SF (match_operand:SF 1 "register_operand" "0")))]
 ""
 "*{ extendReg(operands[0]);
     return \"neg  %0\";
}")

(define_insn "negdf2"
 [(set (match_operand:DF 0 "register_operand" "=u")
   (neg:DF (match_operand:DF 1 "register_operand" "0")))]
 ""
 "*{ extendReg(operands[0]);
     return \"neg  %0\";
}")

(define_insn "abssf2"
 [(set (match_operand:SF 0 "register_operand" "=l")
   (abs:SF (match_operand:SF 1 "register_operand" "0")))]
 ""
 "*{ extendReg(operands[0]);
     return \"abs  %0\";
}")

(define_insn "absdf2"
 [(set (match_operand:DF 0 "register_operand" "=u")
   (abs:DF (match_operand:DF 1 "register_operand" "0")))]
 ""
 "*{ extendReg(operands[0]);
     return \"abs  %0\";
}")

(define_insn "one_cmplqi2"
 [(set (match_operand:QI 0 "register_operand" "=l")
   (not:QI (match_operand:QI 1 "register_operand" "0")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"not  %0\";
}")

(define_insn "one_cmplhi2"
 [(set (match_operand:HI 0 "register_operand" "=l")
   (not:HI (match_operand:HI 1 "register_operand" "0")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"not  %0\";
}")

(define_insn "one_cmplsi2"
 [(set (match_operand:SI 0 "register_operand" "=l")
   (not:SI (match_operand:SI 1 "register_operand" "0")))]
 ""
 "*{ regValid[REGNO(operands[0])]=0;
     return \"not  %0\";
}")

; This is one of the imponderables of the compiler...it can actually
; generate an instruction of the form 'cmp A,A'.  We deal with this
; in NOTICE_UPDATE_CC by setting CC_ALWAYS_TRUE and by using this RTL
; code to not output anything.
(define_insn ""
 [(set (cc0) (compare (match_operand 0 "register_operand" "")
					  (match_dup 0)))]
 ""
 ";cmp reg,reg")

(define_insn "cmpqi"
 [(set (cc0) (compare (match_operand:QI 0 "register_operand" "l")
                      (match_operand:QI 1 "register_operand" "C")))]
 ""
 "*{ extendReg(operands[0]); extendReg(operands[1]);
     return \"cmp  %1,%0\";
}")

(define_insn "cmphi"
 [(set (cc0) (compare (match_operand:HI 0 "register_operand" "l")
                      (match_operand:HI 1 "register_operand" "C")))]
 ""
 "*{ extendReg(operands[0]); extendReg(operands[1]);
     return \"cmp  %1,%0\";
}")

(define_insn "cmpsi"
 [(set (cc0) (compare (match_operand:SI 0 "register_operand" "l")
                      (match_operand:SI 1 "register_operand" "C")))]
 ""
 "*{ extendReg(operands[0]); extendReg(operands[1]);
     return \"cmp  %1,%0\";
}")

(define_insn "cmpdi"
 [(set (cc0) (compare (match_operand:DI 0 "register_operand" "u")
                      (match_operand:DI 1 "register_operand" "u")))]
 ""
 "*{ extendReg(operands[0]); extendReg(operands[1]);
     return \"cmp  %1,%0\";
}")

(define_insn "cmpsf"
 [(set (cc0) (compare (match_operand:SF 0 "register_operand" "l")
                      (match_operand:SF 1 "register_operand" "C")))]
 ""
 "*{ extendReg(operands[0]); extendReg(operands[1]);
     return \"cmp  %1,%0\";
}")

(define_insn "cmpdf"
 [(set (cc0) (compare (match_operand:DF 0 "register_operand" "u")
                      (match_operand:DF 1 "register_operand" "u")))]
 ""
 "*{ extendReg(operands[0]); extendReg(operands[1]);
     return \"cmp  %1,%0\";
}")

(define_insn "tstqi"
 [(set (cc0) (match_operand:QI 0 "register_operand" "l"))]
 ""
 "*{ extendReg(operands[0]);
     return \"tst  %0\";
}")

(define_insn "tsthi"
 [(set (cc0) (match_operand:HI 0 "register_operand" "l"))]
 ""
 "*{ extendReg(operands[0]);
     return \"tst  %0\";
}")

(define_insn "tstsi"
 [(set (cc0) (match_operand:SI 0 "register_operand" "l"))]
 ""
 "*{ extendReg(operands[0]);
     return \"tst  %0\";
}")

(define_insn "tstdi"
 [(set (cc0) (match_operand:DI 0 "register_operand" "u"))]
 ""
 "*{ extendReg(operands[0]);
     return \"tst  %0\";
}")

(define_insn "tstsf"
 [(set (cc0) (match_operand:SF 0 "register_operand" "l"))]
 ""
 "*{ extendReg(operands[0]);
     return \"tst  %0\";
}")

(define_insn "tstdf"
 [(set (cc0) (match_operand:DF 0 "register_operand" "u"))]
 ""
 "*{ extendReg(operands[0]);
     return \"tst  %0\";
}")

(define_insn "floatqisf2"
 [(set (match_operand:SF 0 "register_operand" "=C")
   (float:SF (match_operand:QI 1 "register_operand" "0")))]
 ""
 "; QI->SF")

(define_insn "floathisf2"
 [(set (match_operand:SF 0 "register_operand" "=C")
   (float:SF (match_operand:HI 1 "register_operand" "0")))]
 ""
 "; HI->SF")

(define_insn "floatsisf2"
 [(set (match_operand:SF 0 "register_operand" "=C")
   (float:SF (match_operand:SI 1 "register_operand" "0")))]
 ""
 "; SI->SF")

(define_insn "floatunsqisf2"
 [(set (match_operand:SF 0 "register_operand" "=C")
   (unsigned_float:SF (match_operand:QI 1 "register_operand" "0")))]
 ""
 "; UQI->SF")

(define_insn "floatunshisf2"
 [(set (match_operand:SF 0 "register_operand" "=C")
   (unsigned_float:SF (match_operand:HI 1 "register_operand" "0")))]
 ""
 "; UHI->SF")

(define_insn "floatunssisf2"
 [(set (match_operand:SF 0 "register_operand" "=C")
   (unsigned_float:SF (match_operand:SI 1 "register_operand" "0")))]
 ""
 "; USI->SF")

(define_insn "floatqidf2"
 [(set (match_operand:DF 0 "register_operand" "=A")
   (float:DF (match_operand:QI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0]=1;
	if (reg0 < 4) {		/* Destination is X or Y */
		if (reg0 != reg1) output_asm_insn(\"move %w1,%y0\", operands);
		return \"move #0,%x0\";
	}
    return \"move %w1,%0\";
}")

(define_insn "floathidf2"
 [(set (match_operand:DF 0 "register_operand" "=A")
   (float:DF (match_operand:HI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0] = 1;
	if (reg0 < 4) {		/* Destination is X or Y */
		if (reg0 != reg1) output_asm_insn(\"move %w1,%y0\", operands);
		return \"move #0,%x0\";
	}
    return \"move %w1,%0\";
}")

(define_insn "floatsidf2"
 [(set (match_operand:DF 0 "register_operand" "=A")
   (float:DF (match_operand:SI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0] = 1;
	if (reg0 < 4) {		/* Destination is X or Y */
		if (reg0 != reg1) output_asm_insn(\"move %w1,%y0\", operands);
		return \"move #0,%x0\";
	}
    return \"move %w1,%0\";
}")

(define_insn "floatunsqidf2"
 [(set (match_operand:DF 0 "register_operand" "=A")
   (unsigned_float:DF (match_operand:QI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0]=1;
	if (reg0 < 4) {		/* Destination is X or Y */
		if (reg0 != reg1) output_asm_insn(\"move %w1,%y0\", operands);
		return \"move #0,%x0\";
	}
    return \"move %w1,%0\";
}")

(define_insn "floatunshidf2"
 [(set (match_operand:DF 0 "register_operand" "=A")
   (unsigned_float:DF (match_operand:HI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0]=1;
	if (reg0 < 4) {		/* Destination is X or Y */
		if (reg0 != reg1) output_asm_insn(\"move %w1,%y0\", operands);
		return \"move #0,%x0\";
	}
    return \"move %w1,%0\";
}")

(define_insn "floatunssidf2"
 [(set (match_operand:DF 0 "register_operand" "=A")
   (unsigned_float:DF (match_operand:SI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0]=1;
	if (reg0 < 4) {		/* Destination is X or Y */
		if (reg0 != reg1) output_asm_insn(\"move %w1,%y0\", operands);
		return \"move #0,%x0\";
	}
    return \"move %w1,%0\";
}")

(define_insn "fixsfsi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (fix:SI (match_operand:SF 1 "register_operand" "0")))]
 ""
 ";SF->SI")

(define_expand "fix_truncsfsi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (fix:SI (match_operand:SF 1 "register_operand" "0")))]
 ""
 "")

(define_expand "fixuns_truncsfsi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (fix:SI (match_operand:SF 1 "register_operand" "0")))]
 ""
 "")

(define_insn "fixdfsi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (fix:SI (match_operand:DF 1 "register_operand" "A")))]
 ""
 "*
{
	int reg0 = REGNO(operands[0]);
	int reg1 = REGNO(operands[1]);

	if (reg0 != reg1) {
        regValid[reg0]=1;
        return \"move %y1,%0\";
	} else {
		return \";DF->SI\";
	}
}")

(define_insn "fixdfdi2"
 [(set (match_operand:DI 0 "register_operand" "=A")
   (fix:DI (match_operand:DF 1 "register_operand" "0")))]
 ""
 ";DF->DI")

(define_expand "fix_truncdfdi2"
 [(set (match_operand:DI 0 "register_operand" "=A")
   (fix:DI (match_operand:DF 1 "register_operand" "0")))]
 ""
 "")

(define_expand "fixuns_truncdfdi2"
 [(set (match_operand:DI 0 "register_operand" "=A")
   (fix:DI (match_operand:DF 1 "register_operand" "0")))]
 ""
 "")

(define_insn "fixunssfsi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (unsigned_fix:SI (match_operand:SF 1 "register_operand" "0")))]
 ""
 ";USF->SI")

(define_insn "fixunsdfsi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (unsigned_fix:SI (match_operand:DF 1 "register_operand" "A")))]
 ""
 "*
{
	if (REGNO(operands[0]) != REGNO(operands[1])) {
        regValid[REGNO(operands[0])]=1;
		return \"move %y1,%0\";
	} else {
		return \";UDF->SI\";
	}
}")

(define_insn "fixunsdfdi2"
 [(set (match_operand:DI 0 "register_operand" "=A")
   (unsigned_fix:DI (match_operand:DF 1 "register_operand" "0")))]
 ""
 ";UDF->DI ")

(define_insn "truncdisi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (truncate:SI (match_operand:DI 1 "register_operand" "A")))]
 ""
 "*
{
	int reg0 = REGNO(operands[0]);
	int reg1 = REGNO(operands[1]);

	if (reg0 < 4) {
		if (reg0 == (reg1+1)) {
			return \";DI->SI\";
		} else {
            regValid[reg0] = 1;
			return \"move %x1,%0\";
		}
	} else {
        regValid[reg0]=1;
		return \"move %x1,%0\";
	}
}")

(define_insn "truncdihi2"
 [(set (match_operand:HI 0 "register_operand" "=C")
   (truncate:SI (match_operand:DI 1 "register_operand" "A")))]
 ""
 "*
{
	int reg0 = REGNO(operands[0]);
	int reg1 = REGNO(operands[1]);

	if (reg0 < 4) {
		if (reg0 == (reg1+1)) {
			return \";DI->HI\";
		} else {
            regValid[reg0]=1;
			return \"move %x1,%0\";
		}
    } else {
        regValid[reg0]=1;
		return \"move %x1,%0\";
	}
}")

(define_insn "truncdiqi2"
 [(set (match_operand:QI 0 "register_operand" "=C")
   (truncate:SI (match_operand:DI 1 "register_operand" "A")))]
 ""
 "*
{
	int reg0 = REGNO(operands[0]);
	int reg1 = REGNO(operands[1]);

	if (reg0 < 4) {
		if (reg0 == (reg1+1)) {
			return \";DI->QI\";
		} else {
            regValid[reg0]=1;
			return \"move %x1,%0\";
		}
    } else {
        regValid[reg0]=1;
		return \"move %x1,%0\";
	}
}")

(define_insn "truncsihi2"
 [(set (match_operand:HI 0 "register_operand" "=C")
   (truncate:HI (match_operand:SI 1 "register_operand" "0")))]
 ""
 ";SI->HI")

(define_insn "trunchiqi2"
 [(set (match_operand:QI 0 "register_operand" "=C")
   (truncate:QI (match_operand:HI 1 "register_operand" "0")))]
 ""
 ";HI->QI")

(define_insn "truncsiqi2"
 [(set (match_operand:QI 0 "register_operand" "=C")
   (truncate:QI (match_operand:SI 1 "register_operand" "0")))]
 ""
 ";SI->QI")

(define_insn "truncdfsf2"
 [(set (match_operand:SF 0 "register_operand" "=C")
   (float_truncate:SF (match_operand:DF 1 "register_operand" "A")))]
 ""
 "*
{
	if (REGNO(operands[0]) != REGNO(operands[1])) {
        regValid[REGNO(operands[0])]=1;
        return \"move %y1,%0\";
	} else {
		return \";DF->SF\";
	}
}")

(define_insn "extendqihi2"
 [(set (match_operand:HI 0 "register_operand" "=C")
   (sign_extend:HI (match_operand:QI 1 "register_operand" "0")))]
 ""
 ";QI->HI")

(define_insn "extendhisi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (sign_extend:SI (match_operand:HI 1 "register_operand" "0")))]
 ""
 ";HI->SI")

(define_insn "extendqisi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (sign_extend:SI (match_operand:QI 1 "register_operand" "0")))]
 ""
 ";QI->SI")

(define_insn "extendsidi2"
 [(set (match_operand:DI 0 "register_operand" "=A")
   (sign_extend:DI (match_operand:SI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0 = REGNO(operands[0]);
	int reg1 = REGNO(operands[1]);

    regValid[reg0]=1;
	if (reg0 < 4) {
		if (reg0 != (reg1-1)) output_asm_insn(\"move %w1,%x0\", operands);
		return \"move #0,%y0\";
	}
	if (reg0 != reg1) {
		return \"clr  %0\;move %w1,%x0\";
	}
	return \"move %0,%x0\;move #0,%y0\;move #0,%z0\";
}")

(define_insn "zero_extendqihi2"
 [(set (match_operand:HI 0 "register_operand" "=C")
   (zero_extend:HI (match_operand:QI 1 "register_operand" "0")))]
 ""
 ";UQI->HI")

(define_insn "zero_extendhisi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (zero_extend:SI (match_operand:HI 1 "register_operand" "0")))]
 ""
 ";UHI->SI")

(define_insn "zero_extendqisi2"
 [(set (match_operand:SI 0 "register_operand" "=C")
   (zero_extend:SI (match_operand:QI 1 "register_operand" "0")))]
 ""
 ";UQI->SI")

(define_insn "zero_extendsidi2"
 [(set (match_operand:DI 0 "register_operand" "=A")
   (zero_extend:DI (match_operand:SI 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0]=1;
	if (reg0 < 4) {	/* Is this X1 or Y1 */
		if (reg0 != (reg1-1)) {
			output_asm_insn(\"move %w1,%x0\", operands);
		}
		return \"move #0,%y0\";
	} else {	/* This is A or B */
		if (reg1 != reg0) {
			return \"clr  %0\;move %w1,%x0\";
		}
			/* From A/B to itself */
		output_asm_insn(\"move %0,%x0\", operands);
		output_asm_insn(\"move #0,%y0\", operands);
		return \"move #0,%z0\";
	}
}")

(define_insn "extendsfdf2"
 [(set (match_operand:DF 0 "register_operand" "=A")
   (float_extend:DF (match_operand:SF 1 "register_operand" "C")))]
 ""
 "*
{
	int reg0=REGNO(operands[0]);
	int reg1=REGNO(operands[1]);

    regValid[reg0]=1;
	if (reg0 < 4) {	/* Is this X1 or Y1 */
		if (reg0 == reg1) {
			return \"move #0,%x0\";
		} else {
			return \"move %1,%y0\;move #0,%x0\";
		}
	} else {	/* This is A or B */
		if (reg0 != reg1) {
			return \"move %1,%0\";
		} else {
            return \"move %w1,%0\";
		}
	}
}")

(define_insn "beq"
 [(set (pc) (if_then_else (eq (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jeq  %Z0")

(define_insn "bne"
 [(set (pc) (if_then_else (ne (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jne  %Z0")

(define_insn "blt"
 [(set (pc) (if_then_else (lt (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jlt  %Z0")

(define_insn "bgt"
 [(set (pc) (if_then_else (gt (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jgt  %Z0")

(define_insn "bltu"
 [(set (pc) (if_then_else (ltu (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jlt  %Z0")

(define_insn "bgtu"
 [(set (pc) (if_then_else (gtu (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jgt  %Z0")

(define_insn "ble"
 [(set (pc) (if_then_else (le (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jle  %Z0")

(define_insn "bge"
 [(set (pc) (if_then_else (ge (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jge  %Z0")

(define_insn "bleu"
 [(set (pc) (if_then_else (leu (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jle  %Z0")

(define_insn "bgeu"
 [(set (pc) (if_then_else (geu (cc0) (const_int 0))
    (label_ref (match_operand 0 "" ""))
    (pc)))]
 ""
 "jge  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (eq (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jne  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (ne (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jeq  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (lt (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jge  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (gt (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jle  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (ltu (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jge  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (gtu (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jle  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (le (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jgt  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (ge (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jlt  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (leu (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jgt  %Z0")

(define_insn ""
 [(set (pc) (if_then_else (geu (cc0) (const_int 0)) (pc)
    (label_ref (match_operand 0 "" ""))))]
 ""
 "jlt  %Z0")

(define_insn "jump"
 [(set (pc) (label_ref (match_operand 0 "" "")))]
 ""
 "jmp  %Z0")

(define_insn "call"
 [(call (match_operand    0 "general_operand" "o")
        (match_operand:SI 1 "general_operand" "rmi"))]
 ""
 "jsr  %Z0")

;
; Subroutines return values in A only
;
(define_insn "call_value"
 [(set (match_operand 0 "register_operand" "=u")
    (call (match_operand 1 "general_operand" "o")
          (match_operand:SI 2 "general_operand" "rmi")))]
 ""
 "jsr  %Z1")

(define_insn "nop"
 [(const_int 0)]
 ""
 "nop")

; Motorola's assembler won't put in a NOP before a JMP (R0) if R0 was
; just previously loaded.  This is because this sequence may actually
; be desirable sometimes, as it is in the bootstrap code. So, we have
; to assume the worst.
(define_insn "tablejump"
 [(set (pc) (match_operand 0 "register_operand" "a"))
  (use (label_ref (match_operand 1 "" "")))]
 ""
 "nop\;jmp  (%Z0)")

